/* 
* Blowhole (Game Boy Advance puzzler) - collision_specs.c
* Description: Definitions, global variables, and functions pertaining
*    to collision detection
* License: Copyright (C) 2002 Lance Legan, Something Screwy Productions;
*    This file is part of Blowhole.
*
*    Blowhole is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Blowhole is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Blowhole; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*    The GNU GPL can be found online at: http://www.gnu.org/licenses/
*    Blowhole can be found online at: http://www.somethingscrewy.com/
*/

//Dependencies: include gba.h, sprites.h, sprites_specs.c before this file

#ifndef COLLISION_SPECS_C
#define COLLISION_SPECS_C

//definitions:
typedef enum CollisionType { WATER, ROCKS, NET, PEARL, FISH } CollisionType;


//global variables:
u8 collisionmap[160][120];//__attribute__((section(".ewram"))) = {0};
                          //collisionmap[][] can be put in ewram if
                          //desired, but was not necessary


//functions (with an explanation preceding each):
/*
* RenderCollisionMap(u8[][]) takes a level in the Blowhole standard level
* format (see levels_data.h) and makes a pixel by pixel collision map
* from it.  The collisionmap[][] is 120x160 (even though the screen is 
* 240x160) because the map is "compressed" horizontally by using 4 bits per 
* pixel instead of 8 bits.  The data for each pixel is read by bitwise
* ANDing the 8 bit number with an appropriate bit mask.  This function takes
* too long to execute, and I'm sure can be optimized in some way, or
* a totally different collision scheme could be used.  One idea is to render
* a smaller collision map only for the area of the map where the sprite
* is currently located.  I stuck with the slower method because the 
* implementation is easier, and the delay gives a better animation speed for 
* the moving schools of fish when the collision map is rerendered each frame.  
* Because of the way the GBA accesses memory, the function executes noticibly 
* slower if the counter variables are changed to u16.  Using u8 and u32 both 
* seemed to give about the same results.
*/
void RenderCollisionMap(u8 level[10][15])
{
   u8 i, j, x, y;

   for (y = 0; y < 10; y++) {
      for (x = 0; x < 15; x++) {             //for each level block,
         for (j = 0; j < 16; j++) {
            for (i = 0; i < 8; i++) {        //write a corresponding block size 8x16
               collisionmap[y * 16 + j][x * 8 + i] = (level[y][x] << 4) + level[y][x];
            }
         }
      }
   }
}

/*
* TestCollision() uses the blowhole sprite members and a level's collision 
* map (generated by RenderCollisionMap(u8[][])) to determine what the sprite
* is colliding with.  Various numbers in this function can be tweaked to
* adjust the collision detection, but I suggest they not be touched.
* I have only commented the first case because the rest of the cases 
* follow the same idea.
*/
int TestCollision()
{
   u16 x, y, xpos, ypos, collision;

      xpos = blowhole.x;
      ypos = blowhole.y;

   if (!blowhole.flipped) {
      switch(blowhole.direction) {
         case UP:                         //sprite is moving up
            if (blowhole.x > 224)         //don't allow vertical movment...
               return true;               //when off screen horizontally
            if (blowhole.y == 0 || blowhole.y > 143)
               ypos = 143;                //adjust position for wrapping around screen
            for (y = 0; y < 9; y++) {     //check collision at current position...
               for (x = 1; x < 8; x++) {  //with appropriate part of sprite's collision map
                  collision = collisionmap[ypos + y][(xpos >> 1) + x];
                  if (((blowhole.spriteCollision[y][x] & 0xF0) != 0) && (collision & 0xF0) != 0)
                     return ((collision & 0xF0) >> 4);   //return collision, masking lower 4 bits
                  if (((blowhole.spriteCollision[y][x] & 0x0F) != 0) && (collision & 0x0F) != 0)
                     return collision & 0x0F;            //return collision, masking upper 4 bits
               }
            }
            return WATER;                 //if no collision found, return WATER
            break;
         case DOWN:
            if (blowhole.x > 224)
               return true;
            if (blowhole.y > 143)
               ypos = 0;
            for (y = 11; y < 16; y++) {
               for (x = 1; x < 8; x++) {
                  collision = collisionmap[ypos + y][(xpos >> 1) + x];
                  if (((blowhole.spriteCollision[y][x] & 0xF0) != 0) && (collision & 0xF0) != 0)
                     return ((collision & 0xF0) >> 4);
                  if (((blowhole.spriteCollision[y][x] & 0x0F) != 0) && (collision & 0x0F) != 0)
                     return collision & 0x0F;
               }
            }
            return WATER;
            break;
         case LEFT:
            if (blowhole.y > 144)
               return true;
            if (blowhole.x == 0 || blowhole.x > 224)
               xpos = 224;
            for (y = 5; y < 14; y++) {
               for (x = 0; x < 3; x++) {
                  collision = collisionmap[ypos + y][(xpos >> 1) + x];
                  if (((blowhole.spriteCollision[y][x] & 0xF0) != 0) && (collision & 0xF0) != 0)
                     return ((collision & 0xF0) >> 4);
                  if (((blowhole.spriteCollision[y][x] & 0x0F) != 0) && (collision & 0x0F) != 0)
                     return collision & 0x0F;
               }
            }
            return WATER;
            break;
         default:
            break;
      }
   }
   else {
      switch(blowhole.direction) {
         case UP:
            if (blowhole.x > 224)
               return true;
            if (blowhole.y == 0 || blowhole.y > 143)
               ypos = 143;
            for (y = 0; y < 9; y++) {
               for (x = 1; x < 8; x++) {
                  collision = collisionmap[ypos + y][((xpos + 16) >> 1) - x];
                  if (((blowhole.spriteCollision[y][x] & 0xF0) != 0) && (collision & 0xF0) != 0)
                     return ((collision & 0xF0) >> 4);
                  if (((blowhole.spriteCollision[y][x] & 0x0F) != 0) && (collision & 0x0F) != 0)
                     return collision & 0x0F;
               }
            }
            return WATER;
            break;
         case DOWN:
            if (blowhole.x > 224)
               return true;
            if (blowhole.y > 143)
               ypos = 0;
            for (y = 11; y < 16; y++) {
               for (x = 1; x < 8; x++) {
                  collision = collisionmap[ypos + y][((xpos + 16) >> 1) - x];
                  if (((blowhole.spriteCollision[y][x] & 0xF0) != 0) && (collision & 0xF0) != 0)
                     return ((collision & 0xF0) >> 4);
                  if (((blowhole.spriteCollision[y][x] & 0x0F) != 0) && (collision & 0x0F) != 0)
                     return collision & 0x0F;
               }
            }
            return WATER;
            break;
         case RIGHT:
            if (blowhole.y > 144)
               return true;
            if (blowhole.x > 223)
               xpos = 0;
            for (y = 5; y < 14; y++) {
               for (x = 0; x < 3; x++) {
                  collision = collisionmap[ypos + y][((xpos + 15) >> 1) - x];
                  if (((blowhole.spriteCollision[y][x] & 0xF0) != 0) && (collision & 0xF0) != 0)
                     return ((collision & 0xF0) >> 4);
                  if (((blowhole.spriteCollision[y][x] & 0x0F) != 0) && (collision & 0x0F) != 0)
                     return collision & 0x0F;
               }
            }
            return WATER;
            break;
         default:
            break;
      }
   }
   return WATER;          //if this causes the return, something is FUBAR
}

#endif

